/**
 * Dasu Project
 *
 * File:         DasuNdtModule.java
 * RCS:          $Id: DasuNdtModule.java,v 1.6 2010/06/30 15:07:30 mas939 Exp $
 * Description:  OptimalEdge class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      Oct 15, 2008 at 07:38:30 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2008, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */
package edu.northwestern.dasu.measurement;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedList;

import edu.northwestern.dasu.DasuManager;
import edu.northwestern.dasu.Main;
import edu.northwestern.dasu.MainGeneric;
import edu.northwestern.dasu.drools.FactNdtProbeResult;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.Pair;
import edu.uc.ndt.Tcpbw100;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The DasuNdtModule class manages the set of 
 *         Pings performed by the peer.
 */

public class DasuNdtModule extends DasuProbeModule {	
		
	private class TaskEntry {
		
		public Integer taskId;
		public String server;
		
		public TaskEntry(String server, Integer taskId) {
			this.taskId = taskId;
			this.server = server; 
		}
	}
	
	public Hashtable<Long, FactNdtProbeResult> ndtResultTable;
	private ArrayList<TaskEntry> toDoServerList = new ArrayList<TaskEntry>();
	private static DasuNdtModule self;
	protected static Tcpbw100 ndtTest = new Tcpbw100();
	private static LinkedList<String> servers = new LinkedList<String>();

	public DasuNdtModule() {
		super("DasuNdtModule");
		this.pi = MainGeneric.getPluginInterface();
		ndtResultTable = new Hashtable<Long, FactNdtProbeResult>();
		
		active = true;
	}
	
	public void loadServerList() {
		String myIp = Main.getPublicIpAddress();
		String url = DasuManager.getInstance().getNewRandomDasuUrl() 
			+ "/dasu/ws/ndt.php?ip=" + myIp;
		HashMap<String, String> s = Main.getPropertiesFromServer(url);
		servers =  new LinkedList<String>(s.values());
	}
	
	public synchronized static DasuNdtModule getInstance() {
		if (self == null) {
			self = new DasuNdtModule();
			self.setDaemon(true);
			return self;
		} else {
			return self;
		}
	}

	@Override
	public void run() {
		
		loadServerList();
		
		while (!MainGeneric.isShuttingDown() && isActive()) {
		 
			TaskEntry taskEntry = null;
			synchronized (toDoServerList) {
				while ((toDoServerList.size()==0) && !MainGeneric.isShuttingDown()) {
					try {
						toDoServerList.wait(sleepInterval*1000);
					} catch (InterruptedException e) {
						self = null;
						return;
					}
				}
				
				taskEntry = toDoServerList.remove(0);
				if (DEBUG)
					System.out.println("DasuNdtModule: NDT server " + taskEntry.server );
			}
			
			if (taskEntry.server.equals("HALT"))
				return;

			while (isPause())
				try {
					Thread.sleep(pauseInterval*1000);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			
			FactNdtProbeResult result = (FactNdtProbeResult) probeIp(taskEntry.server, taskEntry.taskId);
			updateNdtResultQueue(result);


	  } // end while active
	
	  self = null;

	}
		
	
	@Override
	public void probeWithPriority(String server, Integer taskId) {
		
		if (server == null)
			server = servers.getFirst();
			
		synchronized (toDoServerList) {
			toDoServerList.add(new TaskEntry(server, taskId));
			toDoServerList.notifyAll();
		}		
	}

	@Override
	public void probe(String server, Integer taskId) {

		if (server == null)
			server = servers.getFirst();

		synchronized (toDoServerList) {
			toDoServerList.add(new TaskEntry(server, taskId));
			toDoServerList.notifyAll();
		}
	}
	
	
	@Override
	protected Object probeIp(String server, Integer taskId) {
	
		if (server == null)
			server = servers.getFirst();
		
		FactNdtProbeResult ret = new FactNdtProbeResult();
		
	    try {
	    	//ret = ndtTest.dottcp("ndt.iupui.ord01.measurement-lab.org");
	    	ret = ndtTest.dottcp(server);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		Statistics.getInstance().addNdtResultDasu(ret);
		return ret;
	}
	
	private void updateNdtResultQueue(FactNdtProbeResult result) {

		if (DEBUG)
			System.out.println("updateNdtResultQueue!!!!");
		
		Long id = System.nanoTime();
		ndtResultTable.put(id, result);
		
		synchronized(AMCoordinatorV2.events) {
			AMCoordinatorV2.events.add(
				new Pair<Long, EventType>(id, EventType.HNDT_RESULT));
			AMCoordinatorV2.events.notifyAll();
		}
	}

	public static LinkedList<String> getServers() {
		return servers;
	}
		
	
	@Override
	public void stopModule() {
		
		setActive(false);
		synchronized(toDoServerList) {
			toDoServerList.add(new TaskEntry("HALT", -1));
			toDoServerList.notify();
		}
		
		self = null;
	}
	
	
}
