/**
 * Dasu Project
 *
 * File:         DasuTraceRouteModuleV2.java
 * RCS:          $Id: DasuTraceRouteModuleV2.java,v 1.4 2010/06/02 02:14:04 mas939 Exp $
 * Description:  DasuTraceRouteModuleV2 class (see below)
 * Author:       Mario Sanchez
 *               Northwestern Systems Research Group
 *               Department of Computer Science
 *               Northwestern University
 * Created:      May 31, 2010 at 5:24:54 PM
 * Language:     Java
 * Package:      edu.northwestern.dasu.measurement
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 2010, Northwestern University, all rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package edu.northwestern.dasu.measurement;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;

import edu.northwestern.dasu.MainGeneric;
import edu.northwestern.dasu.drools.FactTraceResult;
import edu.northwestern.dasu.stats.Statistics;
import edu.northwestern.dasu.util.HashSetCache;
import edu.northwestern.dasu.util.Pair;

/**
 * @author Mario Sanchez &lt;msanchez@u.northwestern.edu&gt;
 * 
 *         The HamecaTraceRouteModule class manages the set of 
 *         trace route measurements performed by the peer.
 */


public class DasuTraceRouteModuleV2 extends DasuProbeModule {	
	
	public Hashtable<Long, FactTraceResult> trResultTable;
	private static DasuTraceRouteModuleV2 self;

    private FactTraceResult tr;
    private HashSetCache<TraceRouteRun> pendingRuns;
    private HashSetCache<TraceRouteRun> pendingRunsPrio;
    private static BufferedReader in;
	private static Process p;
    private Object lossSync = new Object();
    private double lossSum = 0;
    private double lossCount = 0;
	
	public DasuTraceRouteModuleV2() {
		super("DasuTraceRouteModule");
		this.pi = MainGeneric.getPluginInterface();
		self = this;
		active = true;
    	pendingRuns = new HashSetCache<TraceRouteRun>(200);
    	pendingRunsPrio = new HashSetCache<TraceRouteRun>(50);
    	trResultTable = new Hashtable<Long, FactTraceResult>();

	}
	
	public synchronized static DasuTraceRouteModuleV2 getInstance() {
		if (self == null) {
			self = new DasuTraceRouteModuleV2();
			self.setDaemon(true);
			return self;
		} else {
			return self;
		}
	}
	
	@Override
	public void probeWithPriority(String ip, Integer taskId) {
		
		synchronized(pendingRunsPrio){
			pendingRuns.add(new TraceRouteRun(ip));
			pendingRuns.notifyAll();
		}

	}

	@Override
	public void probe(String ip, Integer taskId) {
		
		synchronized(pendingRuns){
			pendingRuns.add(new TraceRouteRun(ip, taskId));
			pendingRuns.notifyAll();
		}
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void run() {

		/** the ip address to trace routes on */
		String dest;
						
		TraceRouteRun trr = null;
		while (!MainGeneric.isShuttingDown() && isActive()){

			synchronized(pendingRuns){
				while (pendingRuns.size()==0 && pendingRunsPrio.size()==0 
						&& !MainGeneric.isShuttingDown()){
					try {
						pendingRuns.wait(sleepInterval*1000);
					} catch (InterruptedException e) {
						setActive(false);
						self = null;
						return;
					}
				}
				
				Iterator<TraceRouteRun> it;
				if (pendingRunsPrio.size()>0){
					
					it = pendingRunsPrio.iterator();
					trr = it.next();
					it.remove();
					if (DEBUG) System.out.println("High priority traceroute to "+trr.ip + " " + trr.taskId);
				} else {
					it = pendingRuns.iterator();
					if (!it.hasNext()){
						continue;
					}
					trr = it.next();
					it.remove();
					if (DEBUG) System.out.println("Normal priority traceroute to "+trr.ip + " " + trr.taskId);
				}
			}
			
			if (trr.ip.equals("HALT")) 
				return;
			
			tr = new FactTraceResult();
			dest = trr.ip;
			
			ArrayList<TraceEntry> data = (ArrayList<TraceEntry>) probeIp(dest, trr.taskId);
			recordInDb(trr, dest, data);
        
		} // end while active

		self = null;
	}
	
	@Override
	protected Object probeIp(String ipAddress, Integer taskId) {

		ArrayList<TraceEntry> data = new ArrayList<TraceEntry>();
		String line = null;
		try {
            
			if (pi.getUtilities().isWindows()) {

                p = Runtime.getRuntime().exec("tracert -d -w 3000 " + ipAddress);
                in = new BufferedReader(new InputStreamReader(
                            p.getInputStream()));              

                while ((line = in.readLine()) != null) {                	
                    // parse data
                	String[] entries = line.split("[\\s]+");
                	if (entries.length<=1) continue;
                	try {
                		Integer.parseInt(entries[1]);
                	} catch (NumberFormatException e){
                		continue;
                	}
                	TraceEntry te = new TraceEntry();
                	int i = 2;
                	int numEntries = 0;
                	while (i < entries.length-1){
                        try{
                		te.rtt[numEntries] = getValue(entries[i]);
                        } catch (NumberFormatException e){
                            // failure during traceroute
                            te.router[te.numRouters++] = getRouterIp(entries[i]);
                            break;
                        }
                		if (te.rtt[numEntries++]<0) i++;
                		else i+=2;
                		if (numEntries==te.rtt.length) break;
                		
                	}
                	
                	te.router[te.numRouters++] = getRouterIp(entries[entries.length-1]);  
                	data.add(te);
                	if (data.size()>=30) break;
                }

                // then destroy
                if (p!=null) p.destroy();         
                
            } else if (pi.getUtilities().isLinux() || pi.getUtilities().isOSX()) {
                p = Runtime.getRuntime().exec("traceroute -n -w 3 " + ipAddress);

                in = new BufferedReader(new InputStreamReader(
                            p.getInputStream()));
                

                while (in!=null && (line = in.readLine()) != null) {                	
                    // parse data
                	
                	String[] entries = line.split("[\\s]+");
                	if (entries.length<=1) continue;
                	int index = 0;
                	while (entries[index].equals("")) index++;
                	try {
                		Integer.parseInt(entries[index]);
                	} catch (NumberFormatException e){
                		// not an entry we care about
                		continue;
                	}
                	TraceEntry te = new TraceEntry();
                	int rttCount = 0;
                	for (int i = index+1; i < entries.length; i++){
                		if (entries[i].contains("*")){
                			te.router[te.numRouters++] = getRouterIp(entries[i]);
                			te.rtt[rttCount++] = -1;
                			//System.out.println("Found asterisk!");
                		}
                		else {
                			if(entries[i].matches("([0-9]{1,3}\\.){3}[0-9]{1,3}")){
                				// if code reaches this point, this is an ip
                				te.router[te.numRouters++] = getRouterIp(entries[i]);
                				//System.out.println("Found router!");
                			}
                			else{
                				// not an ip, so it must be a value
                				try {
	                				te.rtt[rttCount] = Float.parseFloat(entries[i++]);
	                				rttCount++;
                				} catch (NumberFormatException e ){
                					
                					continue; // ignore the garbage
                				}
                				
                				//System.out.println("Found rtt!");
                				// i++ skips the "ms" entry
                				
                			}
                		}
                	}
                 
                	data.add(te);
                	if (data.size()>30) break;
                }

                // then destroy
                if (p!=null) p.destroy();
            }
			
        } catch (IOException e) {
        	if (p!=null)p.destroy();
            //e.printStackTrace();
        } catch (Exception e){
	    	try {
				if (in !=null ) in.close();
			} catch (IOException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
    	
	    	if (p!=null) p.destroy();
	        if (isActive()){
	        	MainGeneric.getReporter().reportError(e);
	        	MainGeneric.getReporter().reportError(new 
	            		RuntimeException("Error input was: \n"+line));
	        }
        }	
        
        return data;

	}
	

	private void recordInDb(TraceRouteRun trr, String dest, ArrayList<TraceEntry> data) {
		
		try {
		
	        if (trr.recordInDB){
	
	            tr.dest = dest;
	            tr.source = MainGeneric.getPublicIpAddress();
	            tr.entries = data;
	            tr.taskId = trr.taskId;
	            
	            updateTrResultQueue(tr);
	            Statistics.getInstance().addTraceRouteResult(tr);
				
	            synchronized(lossSync){
	            	double dropped = 0;
	            	double count = 0;
	            	for (TraceEntry te : data){
	            		int myDropped = 0;
	            		boolean valid = false;
	            		for (int i = 0; i < te.rtt.length; i++){
	            			if (te.rtt[i]>0) valid = true;
	            			else myDropped++;
	            		} // end each entry
	            		if (valid) {
	            			dropped+=myDropped;
	            			count+=te.rtt.length;
	            		}
	            	} // end all tr data
	            	if (count>0){
		            	lossSum+=dropped/count;
		            	lossCount++;
	            	}	
	            } // end sync
	            
	        }
	        if (in!=null) in.close();
	        if (p!=null) p.destroy();
	        
        } catch (IOException e) {
        	if (p!=null)p.destroy();
        }
	}

	private void updateTrResultQueue(FactTraceResult ftr) {

		Long id = System.nanoTime();
		trResultTable.put(id, ftr);
		
		synchronized(AMCoordinatorV2.events) {
			AMCoordinatorV2.events.add(
					new Pair<Long, EventType>(id, EventType.HTRM_RESULT));
			AMCoordinatorV2.events.notifyAll();
		}
	}

	
	private class TraceRouteRun{
		public String ip;
		public boolean recordInDB = true;
		public int taskId = -1;
		
		public TraceRouteRun(String ip){
			this.ip = ip;
		}
		
		public TraceRouteRun(String ip, boolean record){
			this.ip = ip;
			recordInDB = record;
		} 

		public TraceRouteRun(String ip, int taskId) {
			this.ip = ip;
			this.taskId = taskId;
		}
		
		@Override
		public boolean equals(Object obj) {

			return obj!=null && obj instanceof TraceRouteRun && ip.equals(((TraceRouteRun)obj).ip);
		}
		@Override
		public int hashCode() {

			return ip.hashCode();
		}
		
	}
	
	public class TraceEntry {
		public float rtt[] = new float[]{-1f,-1f,-1f};
		public String router[] = new String[3];
		public int numRouters = 0;
	}
	
	private float getValue(String val) {
		if (val.contains("<"))return 0;
		else if (val.contains("*")) return -1;
    	else return Float.parseFloat(val);
	}
	
	public double getLossSample() {
		synchronized(lossSync){
			double toReturn = lossCount==0?0:lossSum/lossCount;
			lossSum = 0;
			lossCount = 0;
			return toReturn;
		}
	}

	private String getRouterIp(String maybeIp) {
		if(maybeIp.matches("([0-9]{1,3}\\.){3}[0-9]{1,3}")){
			// if code reaches this point, this is an ip	
			Statistics.getInstance().addRouterForLookup(maybeIp);
			return maybeIp;
			//System.out.println("Found router!");
		}
		Statistics.getInstance().addRouterForLookup("unknown");
		return "unknown";
	}
		

	@Override
	public void stopModule(){
		
		setActive(false);
				
		synchronized(pendingRuns) {
			pendingRuns.add(new TraceRouteRun("HALT"));
			pendingRuns.notify();
		}

		if (self==null || 
				(self.pendingRuns==null && self.pendingRunsPrio==null)) return;
		if (p!=null){
			p.destroy();
		}
		if (self.pendingRuns!=null){
			synchronized(self.pendingRuns){
				self.pendingRuns.notifyAll();
			}
		}		
		
		self = null;
		p = null;
		pi = null;
	}
	

}
